#version 330 core

/** INPUTS */
in VertexShader {
	vec3 fragpos_worldspace;
	vec3 normal_worldspace;
	vec2 uv;
	vec4 fragpos_lightspace;
} fs_in;


/** UNIFORMS */
uniform vec3 u_CameraPos;
uniform sampler2D u_DepthMap;
uniform samplerCube u_Skybox;


uniform struct ShaderControls {
	bool isTextured;
	vec3 bloomThreshold;
} controls;


uniform struct DirLight {
	vec3 color;
	vec3 direction;         		// Note: from the lightSource towards the fragment!
	vec3 position;					// Note: for shadow mapping only!	
} dirL;


struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
}; 
uniform PointLight pointL[10];
uniform int NR_POINT_LIGHTS;


struct SpotLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
	vec3 direction;
	float innerAngle;
	float outerAngle;
};
uniform SpotLight spotL[10];
uniform int NR_SPOT_LIGHTS;


uniform struct Material {
	vec3 reflectionConstants;      // x = ambient, y = diffuse, z = specular 
	vec3 color;
	float alpha;
	sampler2D textureDiffuse_1;
	sampler2D textureSpecular_1;
	bool hasTexture;
	bool isReflective;
	bool isRefractive;
} material;


/** OUTPUTS */
layout (location = 0) out vec4 fragColor;
layout (location = 1) out vec4 bloomColor;


/** SHADOW CALCULATIONS */

float ShadowCalculation(vec4 fragPosLightSpace, vec3 normal, vec3 lightDir)
{
	// Perform perspective divide
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;

    // Transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;

    // Get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    float closestDepth = texture(u_DepthMap, projCoords.xy).r; 

    // Get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    
	float shadow = 0.0;

	// Shadow bias (offset of surface depth or shadow map)
	// 0.05 - 0.005 range based on the surface's normal and light direction.
	// This results in surfaces that are almost perpendicular to the light source get a smaller bias.
	float bias = max(0.05 * (1.0 - dot(normal, lightDir)), 0.005);

	// Apply percentage-closer filter (PCF) for smoothing shadows
	// Sample surrounding texels and average the result
  	vec2 texelSize = 1.0 / textureSize(u_DepthMap, 0);
    for(int x = -1; x <= 1; ++x)
    {
        for(int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture(u_DepthMap, projCoords.xy + vec2(x, y) * texelSize).r; 
            shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;
        }    
    }
    shadow /= 9.0;
    
    // Keep the shadow at 0.0 when outside the far_plane region of the light's frustum.
    if(projCoords.z > 1.0)
        shadow = 0.0;

    return shadow;
}


/** LIGHT CALCULATIONS */

/* ---------------------------------------- */
// Diffuse, Specular, Attenuation, Phong
/* ---------------------------------------- */

vec3 CalcDiffuse(vec3 lightColor, float diffuseRefConst, vec3 normal, vec3 lightDir) {

	// needs direction FROM the fragment TO the light source
	float cosTheta = max(dot(normal, lightDir), 0.0);
	vec3 diffuse = (lightColor * diffuseRefConst) * cosTheta;
	return diffuse;
}

vec3 CalcSpecular(vec3 lightColor, float specularRefConst, vec3 normal, vec3 lightDir, vec3 eyeDir, float alpha) {

	// needs direction FROM the light source TO the fragment
	vec3 reflectDir = reflect(-lightDir, normal);
    float cosAlpha = max(dot(eyeDir, reflectDir), 0.0);
	float shininess = pow(cosAlpha, alpha);
	vec3 specular = (lightColor * specularRefConst) * shininess;
	return specular;
}

float CalcAttenuation(vec3 lightDir, vec3 attenuation) {

	float distance = length(lightDir);
	float att = 1.0f / (attenuation.x + attenuation.y * distance + attenuation.z * (distance * distance));
	return att;
}

vec3 CalcPhong(vec3 lightColor, vec3 objectColor, vec3 reflectionConstants, vec3 normal, vec3 lightDir, vec3 eyeDir, float alpha, bool shadow) {

	vec3 ambient = lightColor * reflectionConstants.x;
	vec3 diffuse = CalcDiffuse(lightColor, reflectionConstants.y, normal, lightDir);
	vec3 specular = CalcSpecular(lightColor, reflectionConstants.z, normal, lightDir, eyeDir, alpha);

	// Shadows
	if (shadow)	{
    	float shadow = ShadowCalculation(fs_in.fragpos_lightspace, normal, lightDir);

		// Left out ambient color as shadows are rarely completely dark due to light scattering
		// Multiplies with inverse (= how much fragment is not in shadow)
		vec3 shadowAffected = (1.0 - shadow) * (diffuse + specular);

		return (ambient + shadowAffected) * objectColor; 
	} 
	else {
		return (ambient + diffuse + specular) * objectColor;
	}
}


/* ---------------------------- */
// Directional Lightsource
/* ---------------------------- */

vec3 CalcDirLight(DirLight light, Material material, vec3 objectColor, vec3 normal, vec3 eyeDir) {

	// Light direction FROM the fragment TO the light source
	vec3 lightDir = normalize(-light.direction);

	vec3 phong = CalcPhong(light.color, objectColor, material.reflectionConstants, normal, lightDir, eyeDir, material.alpha, true);	
	return phong;
}


/* ---------------------------- */
// Point Lightsource
/* ---------------------------- */

vec3 CalcPointLight(PointLight light, Material material, vec3 objectColor, vec3 normal, vec3 eyeDir, vec3 fragPos) {

	// Light direction FROM the fragment TO the light source
	vec3 lightDir = normalize(light.position - fragPos);

	vec3 phong = CalcPhong(light.color, objectColor, material.reflectionConstants, normal, lightDir, eyeDir, material.alpha, false);
	float attenuation = CalcAttenuation(lightDir, light.attenuation);

	// Add attenuation
	vec3 result = attenuation * phong;
	return result;
}


/* ---------------------------- */
// Spot Lightsource
/* ---------------------------- */

vec3 CalcSpotLight(SpotLight light, Material material, vec3 objectColor, vec3 normal, vec3 eyeDir, vec3 fragPos) {
	
	// Light direction FROM the fragment TO the light source
	vec3 lightDir = normalize(light.position - fragPos);

	vec3 phong = CalcPhong(light.color, objectColor, material.reflectionConstants, normal, lightDir, eyeDir, material.alpha, false);
	float attenuation = CalcAttenuation(lightDir, light.attenuation);

	// SpotLight part
	float theta = dot(lightDir, normalize(-light.direction));				 // the angle between the LightDir and the SpotDir
	float epsilon = light.innerAngle - light.outerAngle;                     // the difference between inner - outerCutoff ==> (theta - outerCutoff) / epsilon
	float intensity = clamp((theta - light.outerAngle) / epsilon, 0.0, 1.0); // clamping avoids if/else statments

	// final
	vec3 result = intensity * attenuation * phong;
	return result;
}



void main()
{
	/** REGULAR Buffer */

	vec3 objectColor;

    // Normalize normal vector
	vec3 normal = normalize(fs_in.normal_worldspace);

	// Eye direction FROM the fragment TO the camera position
	vec3 eyeDir = normalize(u_CameraPos - fs_in.fragpos_worldspace);

	// Define object color
	if (material.hasTexture) {
		objectColor = texture(material.textureDiffuse_1, fs_in.uv).rgb;
	} 
	else {
		objectColor = material.color;
	}


	// Add material properties here
	if (material.isReflective) {
		vec3 R = reflect(-eyeDir, normal);
		vec3 reflectiveColor = texture(u_Skybox, R).rgb;
		objectColor = mix(objectColor, reflectiveColor, 1.0);
	} 

	if (material.isRefractive) {
		float ratio = 1.00 / 1.309;
		vec3 R = refract(-eyeDir, normal, ratio);
		vec3 refractiveColor = texture(u_Skybox, R).rgb;
		objectColor = mix(objectColor, refractiveColor, 0.8);
	} 




	vec3 result = vec3(0.0);

	// Add directional light
	result += CalcDirLight(dirL, material, objectColor, normal, eyeDir);

	// Add point lights
	for (int i = 0; i < NR_POINT_LIGHTS; i++)
	{
		result += CalcPointLight(pointL[i], material, objectColor, normal, eyeDir, fs_in.fragpos_worldspace);
	}
	// Add spot lights
	for (int i = 0; i < NR_SPOT_LIGHTS; i++)
	{
		result += CalcSpotLight(spotL[i], material, objectColor, normal, eyeDir, fs_in.fragpos_worldspace);
	}

	// Final result
	fragColor = vec4(result, 1.0);




	/** BLOOM Buffer */

	// New method
	// float brightness = (fragColor.r * controls.bloomThreshold.x) + (fragColor.g * controls.bloomThreshold.y) + (fragColor.b * controls.bloomThreshold.z);

	// Old method
	// check whether fragment output is higher than threshold, if so output as brightness color
    float brightness = dot(fragColor.rgb, controls.bloomThreshold);

    if (brightness > 1.0)
        bloomColor = vec4(fragColor.rgb, 1.0);
    else
        bloomColor = vec4(0.0, 0.0, 0.0, 1.0);
}